%% Inverse Velocity Kinematics for 2-DOF Planar Arm 
close all; clear; clc;

%% Parameter setting
a1 = 0.5; a2 = 0.5; % unit: m
r = 0.1;  % unit: m, radius for turn around the origin
period = 0.001; % unit: sec
V = 1.0; % unit: m/s, required constant speed  of the end point

elbowup = true; % Elbow-up pose chosen
analytic_method = true; % true: closed-form solution, false: numerical solution
animation = true; % Animation executed

%% Path generation
path_offset = 0.005; % path_offset = 0: fully stretched & singular 

% Path 1 (linear)
% (a1+a2, 0) --> (r, 0) 
path1_x = linspace(a1 + a2 - path_offset, r, (a1 + a2 - r - path_offset) / V / period)';
path1_y = zeros(length(path1_x), 1);

% Path 2 (curve)
% (r, 0) --> curve --> (0, r)
angle_start = 0; angle_end = 90;  % unit: deg
angle_start = angle_start * pi/180; angle_end = angle_end * pi/180;

angular_velocity = V / r;
phi = linspace(angle_start, angle_end, (angle_end - angle_start) / angular_velocity / period)';
path2_x = r * (1 - sin(phi)); path2_y = r * (1 - cos(phi));

% Path 3 (linear)
% (0, r) --> (0, a1 + a2)
path3_y = linspace(r, a1 + a2 - path_offset, (a1 + a2 - r - path_offset) / V / period)';
path3_x = zeros(length(path3_y), 1);

% Entire path
Px = [path1_x ; path2_x(2 : end-1) ; path3_x]; Py = [path1_y ; path2_y(2 : end-1) ; path3_y];

%% Inverse kinematics
% Set DH parameters
if analytic_method == true  % Closed-form solution chosen
    q1 = zeros(length(Px), 1);  q2 = zeros(length(Px), 1);

    for i = 1 : length(Px)
        c2 = (Px(i)^2 + Py(i)^2 - a1^2 - a2^2) / (2*a1*a2); % cos(q2)
        s2p = sqrt((1-c2^2)); % sin(q2): Elbow-down pose 
        s2n = -sqrt((1-c2^2)); % sin(q2): Elbow-up pose
        
        if elbowup == true  
            s2 = s2n; 
        else
            s2 = s2p;
        end
       
        q2(i) = atan2(s2, c2); % Compute q2
        
        s1 = (Py(i) * (a1 + a2*cos(q2(i))) - Px(i)*a2*sin(q2(i)));
        c1 = (Px(i) * (a1 + a2*cos(q2(i))) + Py(i)*a2*sin(q2(i))); 
        q1(i) = atan2(s1, c1); % Compute q1       
    end
else  % Numerical solution
     dhparams = [a1	0   0   0; a2	0   0	0];  % DH parameters: [a alpha d theta]

    % Set rigid body
    body1 = rigidBody('body1');
    joint1 = rigidBodyJoint('joint1', 'revolute');

    setFixedTransform(joint1, dhparams(1,:), 'dh');
    body1.Joint = joint1;

    body2 = rigidBody('body2');
    joint2 = rigidBodyJoint('joint2', 'revolute');

    setFixedTransform(joint2, dhparams(2,:), 'dh');
    body2.Joint = joint2;
    body2.Joint.PositionLimits = [-pi, 0];
    
    robot = rigidBodyTree;
    addBody(robot, body1, 'base');
    addBody(robot, body2, 'body1');

    showdetails(robot);

    ik = inverseKinematics('RigidBodyTree', robot, 'SolverAlgorithm', 'LevenbergMarquardt');    
    ik.SolverParameters.MaxIterations = 50;
    pose_tr = trvec2tform([Px Py zeros(length(Px), 1)]);

    q1 = zeros(length(Px), 1);
    q2 = zeros(length(Px), 1);

    initial_guess = robot.homeConfiguration;
    initial_guess(1).JointPosition = +0.001;
    initial_guess(2).JointPosition = -0.001;

    for i = 1 : length(Px)
        [sol, solInfo] = ik('body2', pose_tr(:, :, i), [0.0000001, 0.0000001, 0.0000001, 10, 10, 10], initial_guess);
        q1(i) = sol(1).JointPosition;
        q2(i) = sol(2).JointPosition;

        initial_guess = sol;
    end
end

% Calculate angular velocity
dq1 = zeros(length(q1) - 1, 1); dq2 = zeros(length(q2) - 1, 1);

for i = 1 : length(dq1)
    dq1(i) = (q1(i+1) - q1(i)) / period; dq2(i) = (q2(i+1) - q2(i)) / period;
end

%% Set Figure Values for positioning of graph windows
Figure_Size = [600, 600];                               % 그래프 사이즈
Figure_1st_Position = [80, 250];                        % 첫번째 그래프 위치
Figure_Position_Increment = [+40 -30];                  % 그래프 간격
Figure_Num = 30;                                        % 그래프 최대 갯수 (넘어도 딱히 상관없음)

Figure_Position = zeros(Figure_Num, 4);

for i = 1 : Figure_Num
   Figure_Position(i, 3:4) = Figure_Size;
   
  if i == 1
      Figure_Position(i, 1:2) = Figure_1st_Position;
  else
      Figure_Position(i, 1) = Figure_Position(i-1, 1) + Figure_Position_Increment(1);
      Figure_Position(i, 2) = Figure_Position(i-1, 2) + Figure_Position_Increment(2);
  end
end

%% Plot graph
% Path
figure('Name', 'Path','NumberTitle','off');
figure_num = 1;
fig = gcf;
set(gcf,'color','w');
fig.Position = Figure_Position(figure_num,:);

hold on;
plot(Px, Py, 'r');
xlabel('x (m)');
ylabel('y (m)');
x_min = min(Px) - (max(Px) - min(Px)) * 0.1;
x_max = max(Px) + (max(Px) - min(Px)) * 0.1;
y_min = min(Py) - (max(Py) - min(Py)) * 0.1;
y_max = max(Py) + (max(Py) - min(Py)) * 0.1;
axis([x_min x_max y_min y_max]);
grid on;

% Joint angle
figure('Name', 'Joint angle','NumberTitle','off');
figure_num = figure_num + 1;
fig = gcf;
set(gcf,'color','w');
fig.Position = Figure_Position(figure_num,:);

x_temp = 0 : length(Px) - 1;

subplot(2, 1, 1);
hold on;
plot(x_temp, q1*180/pi, 'k');
xlabel('time (msec)');
ylabel('joint 1 angle (deg)');
[x_min, x_max, y_min, y_max] = AxisRange([0;length(q1)], q1*180/pi);
axis([x_min x_max y_min y_max]);
grid on;

subplot(2, 1, 2);
hold on;
plot(x_temp, q2 * 180 / pi, 'k');
xlabel('time (msec)');
ylabel('joint 2 angle (deg)');
[x_min, x_max, y_min, y_max] = AxisRange([0;length(q2)], q2*180/pi);
axis([x_min x_max y_min y_max]);
grid on;

% Joint velocity
figure('Name', 'Joint velocity','NumberTitle','off');
figure_num = figure_num + 1;
fig = gcf;
set(gcf,'color','w');
fig.Position = Figure_Position(figure_num,:);

x_temp = 0 : length(dq1) - 1;

subplot(2, 1, 1);
hold on;
plot(x_temp, dq1*180/pi, 'k');
xlabel('Time (msec)');
ylabel('joint 1 velocity (deg/s)');
[x_min, x_max, y_min, y_max] = AxisRange([0;length(dq1)], dq1*180/pi);
axis([x_min x_max y_min y_max]);
grid on;

subplot(2, 1, 2);
hold on;
plot(x_temp, dq2*180/pi, 'k');
xlabel('Time (msec)');
ylabel('joint 2 velocity (deg/s)');
[x_min, x_max, y_min, y_max] = AxisRange([0;length(dq2)], dq2*180/pi);
axis([x_min x_max y_min y_max]);
grid on;

% Combined joint velocity
figure('Name', 'Combined joint velocity','NumberTitle','off');
figure_num = figure_num + 1;
fig = gcf;
set(gcf,'color','w');
fig.Position = Figure_Position(figure_num,:);

x_temp = 0 : length(dq1) - 1;

hold on;
plot(x_temp, dq1*180/pi, 'r', x_temp, dq2*180/pi, 'b');
xlabel('Time (msec)');
ylabel('joint velocity (deg/s)');
legend('Joint 1', 'Joint 2');
[x_min, x_max, y_min, y_max] = AxisRange([0;length(dq1)], [dq1*180/pi, dq2*180/pi]);
axis([x_min x_max y_min y_max]);
grid on;

%% Forward kinematics & animation
if animation == true
    x1 = a1 * cos(q1);
    y1 = a1 * sin(q1);
    x2 = a1 * cos(q1) + a2 * cos(q1+q2);
    y2 = a1 * sin(q1) + a2 * sin(q1+q2);

    figure('Name', 'Forward kinematics','NumberTitle','off');
    figure_num = figure_num + 1;
    fig = gcf;
    set(gcf,'color','w');
    fig.Position = Figure_Position(figure_num,:);

    plot(Px, Py, 'r--');
    xlabel('x (m)');
    ylabel('y (m)');
    x_min = min(Px) - (max(Px) - min(Px)) * 0.1;
    x_max = max(Px) + (max(Px) - min(Px)) * 0.1;
    y_min = min(Py) - (max(Py) - min(Py)) * 0.1;
    y_max = max(Py) + (max(Py) - min(Py)) * 0.1;
    axis([-x_max/1.7 x_max -y_max/1.7 y_max]);
    grid on;
    xlabel('x (m)');
    ylabel('y (m)');

    line1 = line([0, x1(1)], [0, y1(1)], 'color', 'k', 'marker', 'o');
    line2 = line([x1(1), x2(1)], [y1(1), y2(1)], 'color', 'k', 'marker', 'o');

    for i = 2 : length(q1)
        delete(line1);
        delete(line2);

        line1 = line([0, x1(i)], [0, y1(i)], 'color', 'k', 'marker', 'o');
        line2 = line([x1(i), x2(i)], [y1(i), y2(i)], 'color', 'k', 'marker', 'o');
        drawnow;    
    end
end

